/***********************************************************************************

	Projekt:	Ein Betriebsystem fr AVR-8Bit Mikrocontroller
	Datei:		fifo.S
	Autor:		Mark Prediger
	Datum:		08.02.2009
	Copyright:	Mark Prediger, mark_prediger@hotmail.de
	Lizenz:		GNU GPL
	
	Dieses Programm darf frei kopiert und gendert werden, solange
	es nicht kommerziell verwendet wird. Die kommerzielle Nuztung bentigt
	meine Erlaubnis.

	Beschreibung:
		Funktionsdefinitionen fr den Fifo

	Funktionsweise:
		Stellt verschiedene Funktionen fr FIFOs (Ringbuffer) bereit

***********************************************************************************/

#include <avr/io.h>
#include "fifo.h"


.global PushFifo	// void PushFifo(Fifo*, unsigned char);
.func PushFifo
PushFifo:
	movw ZL, 24

	in 0, _SFR_IO_ADDR(SREG)
	cli

	ldd XL, Z+4	// X = fifo->p_in
	ldd XH, Z+5

	st -X, 22	// *(--fifo->p_in) = c;

	ldd 22, Z+0	// 22:23 = fifo->buffer
	ldd 23, Z+1

	cp 22, XL
	cpc 23, XH
	brcs pushfifo_no_underflow	// if(fifo->p_in<=fifo->buffer)
								//		fifo->p_in = fifo->buffer+fifo->size;
	movw XL, 22	// X = fifo->buffer

	ldd 20, Z+2	// 20:21 = fifo->size
	ldd 21, Z+3

	add XL, 20	// X = fifo->buffer + fifo->size
	adc XH, 21

pushfifo_no_underflow:
	std Z+4, XL	// fifo->p_in = X
	std Z+5, XH

	ldd 24, Z+6	// 24:25 = fifo->p_out
	ldd 25, Z+7

	cp XL, 24
	cpc XH, 25
	brne pushfifo_done	// if(fifo->p_in==fifo->p_out)

	sbiw 24, 1	// fifo->p_out--;

	cp 22, 24
	cpc 23, 25
	brcs pushfifo_no_out_undeflow	// if(fifo->p_out<fifo->buffer) fifo->p_out = fifo->buffer+fifo->size;

	movw 24, 22	// 24:25 fifo->buffer

	add 24, 20	// 24:25 = fifo->buffer + fifo->size
	adc 25, 21

pushfifo_no_out_undeflow:
	std Z+6, 24	// fifo->p_out = 24:25
	std Z+7, 25

pushfifo_done:
	out _SFR_IO_ADDR(SREG), 0
	ret
.endfunc

.global PopFifo	// unsigned char PopFifo(Fifo*);
.func PopFifo
PopFifo:
	in 0, _SFR_IO_ADDR(SREG)
	movw ZL, 24

popfifo_wait:
	cli
	ldd 22, Z+4	// 22:23 = fifo->p_in
	ldd 23, Z+5
	ldd XL, Z+6	// X = fifo->p_out
	ldd XH, Z+7

	cp XL, 22
	cpc XH, 23
	brne popfifo_got_byte		// while(fifo->p_in == fifo->p_out) SwitchToThread()
	out _SFR_IO_ADDR(SREG), 0
	call SwitchToThread
	rjmp popfifo_wait

popfifo_got_byte:
	ld 24, -X
	clr 25

	ldd 22, Z+0	// 22:23 = fifo->buffer
	ldd 23, Z+1

	cp 22, XL
	cpc 23, XH
	brcs popfifo_no_underflow	// if(fifo->p_in<=fifo->buffer)	fifo->p_in = fifo->buffer+fifo->size;

	movw XL, 22	// X = fifo->buffer

	ldd 20, Z+2	// 20:21 = fifo->size
	ldd 21, Z+3

	add XL, 20	// X = fifo->buffer + fifo->size
	adc XH, 21

popfifo_no_underflow:
	std Z+6, XL	// fifo->p_out = fifo->buffer + fifo->size
	std Z+7, XH

	out _SFR_IO_ADDR(SREG), 0
	ret
.endfunc

.global BytesInFifo	// int BytesInFifo(Fifo*);
.func BytesInFifo
BytesInFifo:
	in 0, _SFR_IO_ADDR(SREG)
	cli
	ldd 22, Z+4	// 22:23 = fifo->p_in
	ldd 23, Z+5
	ldd 24, Z+6	// 24:25 = fifo->p_out
	ldd 25, Z+7
	out _SFR_IO_ADDR(SREG), 0

	cp 24, 22
	cpc 25, 23
	brcc bytesinfifo_not_bigger

	// if(p_in>p_out) return p_out-p_in+size; else return return p_out-p_in;

	ldd XL, Z+2	// X = fifo->size
	ldd XH, Z+3

	add 24, XL
	adc 25, XH

bytesinfifo_not_bigger:
	sub 24, 22
	sbc 25, 23
	ret
.endfunc
