/***********************************************************************************

	Projekt:	Ein Betriebsystem fr AVR-8Bit Mikrocontroller
	Datei:		message.S
	Autor:		Mark Prediger
	Datum:		08.02.2009
	Copyright:	Mark Prediger, mark_prediger@hotmail.de
	Lizenz:		GNU GPL
	
	Dieses Programm darf frei kopiert und gendert werden, solange
	es nicht kommerziell verwendet wird. Die kommerzielle Nuztung bentigt
	meine Erlaubnis.

	Beschreibung:
		Funktionsdefinitionen fr die Verwendung von Nachrichen

	Funktionsweise:
		Eine Nachricht (Message) besteht aus einem Funktionspointer, 
		der Nachricht selbst (int msg) und einem Parameter (param).
		
		Mann kann Nachrichten auf den Nachrichten-FIFO legen oder 
		vom FIFO abhohlen. Wenn beim legen ein NULL-Pointer fr die 
		Funktion angegeben wird, wird die Standard-Prozedur des
		Systems genommen. Sollte debei der Nachrichten-FIFO voll sein,
		wird die letzte Nachricht entfernt. SendMessage(), SendMessageByPointer()
		und SendFastMessage knnen auch in einer ISR aufgerufen werden.
		
		Die Standard-Prozedur ist zu Beginn auf SystemDefProc gestellt, kann
		aber durch SetSystemDefProc jederzeit durch eine eigene ersetzt werden.

***********************************************************************************/

#include <avr/io.h>
#include "message.h"

.extern message_fifo, comm message_in_index, message_out_index


.global SendMessage	// void SendMessage(void(*)(int, int), int, int);
.func SendMessage
SendMessage:
	cp 24, 1
	cpc 25, 1
	brne send_not_null // if(pfunc==NULL) pfunc = system_def_proc

	lds 24, system_def_proc
	lds 25, system_def_proc+1

send_not_null:
	in 19, _SFR_IO_ADDR(SREG)
	cli

	lds ZL, message_in_index
	lds ZH, message_in_index+1
	st Z+, 24
	st Z+, 25
	st Z+, 22
	st Z+, 23
	st Z+, 20
	st Z+, 21

	ldi 18, hi8(message_fifo+MESSAGE_FIFO_SIZE*MESSAGE_STRUCT_SIZE)
	cpi ZL, lo8(message_fifo+MESSAGE_FIFO_SIZE*MESSAGE_STRUCT_SIZE)
	cpc ZH, 18
	brcs message_no_overflow
	ldi ZL, lo8(message_fifo)
	ldi ZH, hi8(message_fifo)
message_no_overflow:
	sts message_in_index, ZL
	sts message_in_index+1, ZH

	lds XL, message_out_index
	lds XH, message_out_index+1
	
	cp ZL, XL
	cpc ZH, XH
	brne message_not_full		// if(in_index==out_index) outindex++

	adiw XL, MESSAGE_STRUCT_SIZE
	cpi XL, lo8(message_fifo+MESSAGE_FIFO_SIZE*MESSAGE_STRUCT_SIZE)
	cpc XH, 18
	brcs message_no_out_overflow

	ldi XL, lo8(message_fifo)
	ldi XH, hi8(message_fifo)

message_no_out_overflow:
	sts message_out_index, XL
	sts message_out_index+1, XH

message_not_full:
	out _SFR_IO_ADDR(SREG), 19
	ret
.endfunc

.global SendMessageByPointer	// void SendMessageByPointer(Message*);
.func SendMessageByPointer
SendMessageByPointer:
	in 19, _SFR_IO_ADDR(SREG)
	cli

	movw XL, 24
	lds ZL, message_in_index
	lds ZH, message_in_index+1

	ld 24, X+			// r24:r25 = msg->pfunc
	ld 25, X+

	cp 24, 1
	cpc 25, 1
	brne send_p_not_null	// if(pfunc==NULL) pfunc = system_def_proc

	lds 24, system_def_proc
	lds 25, system_def_proc+1

send_p_not_null:
	st Z+, 24
	st Z+, 25

	ld 18, X+
	st Z+, 18
	ld 18, X+
	st Z+, 18
	ld 18, X+
	st Z+, 18
	ld 18, X+
	st Z+, 18

	ldi 18, hi8(message_fifo+MESSAGE_FIFO_SIZE*MESSAGE_STRUCT_SIZE)
	cpi ZL, lo8(message_fifo+MESSAGE_FIFO_SIZE*MESSAGE_STRUCT_SIZE)
	cpc ZH, 18
	brcs message_p_no_overflow
	ldi ZL, lo8(message_fifo)
	ldi ZH, hi8(message_fifo)
message_p_no_overflow:
	sts message_in_index, ZL
	sts message_in_index+1, ZH

	lds XL, message_out_index
	lds XH, message_out_index+1

	cp ZL, XL
	cpc ZH, XH
	brne message_p_not_full		// if(in_index==out_index) outindex++

	adiw XL, MESSAGE_STRUCT_SIZE
	cpi XL, lo8(message_fifo+MESSAGE_FIFO_SIZE*MESSAGE_STRUCT_SIZE)
	cpc XH, 18
	brcs message_p_no_out_overflow

	ldi XL, lo8(message_fifo)
	ldi XH, hi8(message_fifo)

message_p_no_out_overflow:
	sts message_out_index, XL
	sts message_out_index+1, XH

message_p_not_full:
	out _SFR_IO_ADDR(SREG), 19
	ret
.endfunc


.global SendFastMessage	// void SendFastMessage(void(*)(int, int), int, int);
.func SendFastMessage
SendFastMessage:
	sbiw 24, 0
	brne send_fast_not_null // if(pfunc==NULL) pfunc = system_def_proc

	lds 24, system_def_proc
	lds 25, system_def_proc+1

send_fast_not_null:
	in 19, _SFR_IO_ADDR(SREG)
	cli

	lds ZL, message_out_index	// Z = message_out_index-1
	lds ZH, message_out_index+1
	sbiw ZL, MESSAGE_STRUCT_SIZE

	ldi 18, hi8(message_fifo)	// auf underflow prfen
	cpi ZL, lo8(message_fifo)
	cpc ZH, 18
	brcc send_fast_no_undeflow
	ldi ZL, lo8(message_fifo+MESSAGE_FIFO_SIZE*MESSAGE_STRUCT_SIZE-MESSAGE_STRUCT_SIZE)
	ldi ZH, hi8(message_fifo+MESSAGE_FIFO_SIZE*MESSAGE_STRUCT_SIZE-MESSAGE_STRUCT_SIZE)
send_fast_no_undeflow:

	lds XL, message_in_index
	lds XH, message_in_index+1

	cp ZL, XL
	cpc ZH, XH
	brne send_fast_not_full		// if(fifo voll) berschreibe Nachricht an out

	lds ZL, message_out_index
	lds ZH, message_out_index+1
	rjmp send_fast_full

send_fast_not_full:
	sts message_out_index, ZL
	sts message_out_index+1, ZH

send_fast_full:
	st Z+, 24
	st Z+, 25
	st Z+, 22
	st Z+, 23
	st Z+, 20
	st Z+, 21

	out _SFR_IO_ADDR(SREG), 19
	ret
.endfunc


.global GetMessage	// short GetMessage(Message*);
.func GetMessage
GetMessage:

get_wait_message:				// warten bis Message im FIFO
	cli
	lds XL, message_in_index
	lds XH, message_in_index+1
	lds ZL, message_out_index
	lds ZH, message_out_index+1
	cp XL, ZL
	cpc XH, ZH

	brne get_got_message
	sei
	call SwitchToThread
	rjmp get_wait_message

get_got_message:
	movw XL, 24

	ld 18, Z+
	st X+, 18
	ld 18, Z+
	st X+, 18

	ld 24, Z+		// msg fr return
	st X+, 24
	ld 25, Z+
	st X+, 25

	ld 18, Z+
	st X+, 18
	ld 18, Z+
	st X+, 18

	ldi 18, hi8(message_fifo+MESSAGE_FIFO_SIZE*MESSAGE_STRUCT_SIZE)
	cpi ZL, lo8(message_fifo+MESSAGE_FIFO_SIZE*MESSAGE_STRUCT_SIZE)
	cpc ZH, 18
	brcs get_no_overflow
	ldi ZL, lo8(message_fifo)
	ldi ZH, hi8(message_fifo)
get_no_overflow:
	sts message_out_index, ZL
	sts message_out_index+1, ZH

	sei
	ret
.endfunc

.global DispatchNextMessage	// short DispatchNextMessage();
.func DispatchNextMessage
DispatchNextMessage:

dispatch_wait_message:				// warten bis Message im FIFO
	cli
	lds ZL, message_in_index
	lds ZH, message_in_index+1
	lds XL, message_out_index
	lds XH, message_out_index+1
	cp XL, ZL
	cpc XH, ZH

	brne dispatch_got_message
	sei
	call SwitchToThread
	rjmp dispatch_wait_message

dispatch_got_message:

	ld ZL, X+
	ld ZH, X+
	ld 24, X+
	ld 25, X+
	ld 22, X+
	ld 23, X+

	ldi 18, hi8(message_fifo+MESSAGE_FIFO_SIZE*MESSAGE_STRUCT_SIZE)
	cpi XL, lo8(message_fifo+MESSAGE_FIFO_SIZE*MESSAGE_STRUCT_SIZE)
	cpc XH, 18
	brcs dispatch_no_overflow
	ldi XL, lo8(message_fifo)
	ldi XH, hi8(message_fifo)
dispatch_no_overflow:
	sts message_out_index, XL
	sts message_out_index+1, XH

	sei

	push 24
	push 25
	icall
	pop 25		// return msg
	pop 24

	ret
.endfunc

.global GotMessages	// bool GotMessages();
.func GotMessages
GotMessages:
	cli
	lds XL, message_in_index
	lds XH, message_in_index+1
	lds ZL, message_out_index
	sei
	lds ZH, message_out_index+1

	clr 24
	clr 25

	cp XL, ZL
	cpc XH, ZH
	breq gotmsg_false

	ldi 24, 1
gotmsg_false:
	ret
.endfunc

