/***********************************************************************************
	
	Projekt:	Ein Betriebsystem fr AVR-8Bit Mikrocontroller
	Datei:		message.h
	Autor:		Mark Prediger
	Datum:		08.02.2009
	Copyright:	Mark Prediger, mark_prediger@hotmail.de
	Lizenz:		GNU GPL
	
	Dieses Programm darf frei kopiert und gendert werden, solange
	es nicht kommerziell verwendet wird. Die kommerzielle Nuztung bentigt
	meine Erlaubnis.
	
	Beschreibung: 
		Funktionsdeklarationen und -definitionen fr die Verwendung von Nachrichen
	
	Funktionsweise:
		siehe message.S.
	
***********************************************************************************/

#ifndef MESSAGE_H_INCLUDED
#define MESSAGE_H_INCLUDED

#include <avr/io.h>
#include <avr/interrupt.h>

#define MESSAGE_STRUCT_SIZE	6	// Gre der Message-Struktur
#define MESSAGE_FIFO_SIZE	16	// Gre des Message-FIFOs

// verschiedene Nachrichten 
// Aktuell fr Roboter ARIPS, bitte fr eigene Programme anpassen
#define MSG_QUIT		0		
#define MSG_CREATE		1
#define MSG_TIMER		2
#define MSG_KEY			3
#define MSG_SUB_UART	4
#define MSG_UART		5

// Anfangswert fr amwendungsspezifische Nachrichten
#define MSG_USER_START	0x100

#ifndef __ASSEMBLER__

typedef struct				// Nachricht-Struktur
{
	void(*pfunc)(int16_t, int16_t);	// Callback-Funktion
	int16_t msg;				// Nachricht
	int16_t param;				// Parameter
} Message;

extern volatile Message message_fifo[MESSAGE_FIFO_SIZE];	// FIFO fr die Nachrichten
extern volatile Message *volatile message_in_index, *volatile message_out_index;

extern void(* volatile system_def_proc) (int16_t, int16_t);	// enthlt einen Zeiger auf die aktielle Standard-Prozedur

extern void SystemDefProc(int16_t, int16_t);	// anfngliche Standard-Prozedur

#ifdef __cplusplus
extern "C" {
#endif
/**************************************************************************
	Fgt eine Nachricht ans Ende der Schlange an
**************************************************************************/
void SendMessage(void(*)(int16_t, int16_t), int16_t, int16_t);


/**************************************************************************
	Fgt eine Nachricht ans Ende der Schlange an
**************************************************************************/
void SendMessageByPointer(Message*);


/**************************************************************************
	Fgt eine Nachricht an den Anfang der Schlange an
**************************************************************************/
void SendFastMessage(void(*)(int16_t, int16_t), int16_t, int16_t);


/**************************************************************************
	Holt eine Nachricht vom Anfang der Schlange und gibt msg der Nachricht zurck
**************************************************************************/
int16_t GetMessage(Message*);


/**************************************************************************
	1 wenn Nachricht(en) vorhanden, 0 wenn nicht
**************************************************************************/
int16_t GotMessages();


/**************************************************************************
	Holt eine Nachricht vom Anfang der Schlange, verarbeiten diese und 
	gibt msg der Nachricht zurck
**************************************************************************/
int16_t DispatchNextMessage();	 
#ifdef __cplusplus
}
#endif

/**************************************************************************
	Verarbeitet die Nachricht
**************************************************************************/
inline void DispatchMessage(Message *msg)
{
	msg->pfunc(msg->msg, msg->param);
}

/**************************************************************************
	 berschreibt system_def_proc mit einer neuen Prozedur
**************************************************************************/
inline void SetSystemDefProc(void(*new_def_proc)(int16_t, int16_t))
{
	uint8_t sreg = SREG;
	cli();
	system_def_proc = new_def_proc;
	SREG = sreg;
}

/**************************************************************************
	Gibt dein Zeiger auf die aktuelle Standard-Prozedur zurck
**************************************************************************/
inline void(*GetSystemDefProc())(int16_t, int16_t)
{
	uint8_t sreg = SREG;
	cli();
	void(*def_proc)(int16_t, int16_t) = system_def_proc;
	SREG = sreg;
	return def_proc;
}

/*************************************************************
	Fgt eine Nachricht ans Ende der Schlange an
	 - Fr ISRs, Interrupts werden nicht deaktiviert
	 - Iniline um Pro-/Epilog klein zu halten 
 *************************************************************/
inline void IsrSendMessage(void(*pfunc)(int16_t, int16_t), int16_t msg, int16_t param)
{	
	volatile Message* in = message_in_index;
	in->pfunc = pfunc;
	in->msg = msg;
	in->param = param;
	
	if((++in)>=message_fifo+MESSAGE_FIFO_SIZE)
		in = message_fifo;
	message_in_index = in;
	
	volatile Message* out = message_out_index;
	if(in == out)	// if(fifo voll) letzen eintrag lschen
	{
		if((++out)>=message_fifo+MESSAGE_FIFO_SIZE)
			out = message_fifo;
		message_out_index = out;
	}
}
				 
#endif // __ASSEMBLER__

#endif // MESSAGE_H_INCLUDED
