/***********************************************************************************

	Projekt:	Ein Betriebsystem fr AVR-8Bit Mikrocontroller
	Datei:		mutex.h
	Autor:		Mark Prediger
	Datum:		08.02.2009
	Copyright:	Mark Prediger, mark_prediger@hotmail.de
	Lizenz:		GNU GPL
	
	Dieses Programm darf frei kopiert und gendert werden, solange
	es nicht kommerziell verwendet wird. Die kommerzielle Nuztung bentigt
	meine Erlaubnis.
	
	Beschreibung:
		Funktionsdeklarationen fr die Verwendung von Mutex

	Funktionsweise:
		siehe mutex.cpp

***********************************************************************************/

#ifndef MUTEX_H_INCLUDED
#define MUTEX_H_INCLUDED

#define MUTEX_STATE_UNLOCKED	0
#define MUTEX_STATE_LOCKED		1

#ifndef __ASSEMBLER__

#include "os.h"

struct MutexWaitElement
{
	HTHREAD hthread;
	MutexWaitElement *next;
};

struct Mutex
{
	volatile int8_t state;	// 0 = unlocked, 1 = locked
	MutexWaitElement* volatile wait_queue;
};

/**************************************************************************
	Initialisiert die Mutex-Variable
**************************************************************************/
void InitMutex(Mutex* m, int8_t = MUTEX_STATE_UNLOCKED);

/**************************************************************************
	Wartet, bis die Mutex-Variable frei ist und sperrt sie
**************************************************************************/
void LockMutex(Mutex*);

/**************************************************************************
	Sperrt die Mutex-Variable falls sie frei ist und gibt 0 zurck,
	ansonsten -1
**************************************************************************/
int8_t TryLockMutex(Mutex*);

/**************************************************************************
	Gibt die Mutex-Variable frei
**************************************************************************/
void ReleaseMutex(Mutex* m);

/**************************************************************************
	Wartet, bis die Mutex-Variable frei ist
**************************************************************************/
void WaitMutex(Mutex* m);

#endif // __ASSEMBLER__

#endif // MUTEX_H_INCLUDED
