/***********************************************************************************

	Projekt:	Ein Betriebsystem fr AVR-8Bit Mikrocontroller
	Datei:		timer.S
	Autor:		Mark Prediger
	Datum:		08.02.2009
	Copyright:	Mark Prediger, mark_prediger@hotmail.de
	Lizenz:		GNU GPL
	
	Dieses Programm darf frei kopiert und gendert werden, solange
	es nicht kommerziell verwendet wird. Die kommerzielle Nuztung bentigt
	meine Erlaubnis.

	Beschreibung:
		Funktionsdefinitionen fr die Verwendung von timern

	Funktionsweise:
		Es wird eine angegebene Anzahl an freien Timer-Slots fr die Timer angelegt
		Wenn ein Timer gesetzt wird, wird in den nchsten freien Slot die zu sendende
		Nachricht (->messages.h) und der Timerconter+Maximalwert geschrieben. Falls der
		angegebene Funtkionszeiger NULL ist, wird der durch den Zeiger auf die
		Standrard-Prozedur ersetzt, da ein NULL-Zeiger spter in der ISR eine spezielle
		Bedeutung hat.

		Alternativ kann mann einen Thread aktivieren lassen, nachdem die angegebene
		Zeit abgelaufen ist. Dazu wird der Funktionszeiger der Message auf NULL gesetzt,
		was der ISR des Timers sagt, dass keine Nachricht geschickt sondern ein Thread
		aktiviert werden soll. Der Paramerter msg ist dabei das Handle zum Thread

		Timer-ISR: Bei jedem aktiven Timer(count!=0) wird der Zhler dekrementiert.
		Falls dieser 0 wird, wird entweder eine Nachricht geschickt oder ein Thread
		aktiviert (s.o.) und der Counter mit dem Maximalwert geladen. Falls dieser
		auch 0 ist gilt der Timer als deaktiviert. Standartmig wird der
		System-Timer-IRQ jede 10ms ausgelst, was man durch Umdefinieren von
		SYSTEM_TIMER_TICKS ndern kann (s.u.).

***********************************************************************************/

#include <avr/io.h>
#include "os.h"
#include "timer.h"

.global SetTimer		// HTIMER SetTimer(unsigned int, void(*)(int, int), int, int);
.func SetTimer
SetTimer:
	cp 24, 1
	cpc 25, 1
	breq timer_no_count	// if(count==0) return NULL

	lds ZL, _timerindex
	lds ZH, _timerindex+1

	in 0, _SFR_IO_ADDR(SREG)
	cli

	push 16
	// nach einem freien Slot suchen
	ldi 16, hi8(_timers+(MAX_TIMERS*TIMER_STRUCT_SIZE))
while_timer_thread:
	cpi ZL, lo8(_timers+(MAX_TIMERS*TIMER_STRUCT_SIZE))
	cpc ZH, 16
	brcc timer_thread_no_free_slots
	ldd XL, Z+MESSAGE_STRUCT_SIZE+0
	ldd XH, Z+MESSAGE_STRUCT_SIZE+1
	cp XL, 1
	cpc XH, 1
	breq timer_thread_slot_found		// while(timer->count)timer++
	adiw ZL, TIMER_STRUCT_SIZE
	rjmp while_timer_thread

timer_thread_no_free_slots:
	sts _timerindex, ZL
	out _SFR_IO_ADDR(SREG), 0
	sts _timerindex+1, ZH
	pop 16
	clr 24		// return NULL
	clr 25
timer_no_count:
	ret

timer_thread_slot_found:
	cp 22, 1
	cpc 23, 1
	brne set_timer_not_null

	lds 22, system_def_proc
	lds 23, system_def_proc+1

set_timer_not_null:
	movw XL, ZL							// neuen timerindex speichern
	adiw XL, TIMER_STRUCT_SIZE
	sts _timerindex, XL
	sts _timerindex+1, XH

	std Z+MESSAGE_STRUCT_SIZE+0, 24		// timer->count = count
	std Z+MESSAGE_STRUCT_SIZE+1, 25
	std Z+MESSAGE_STRUCT_SIZE+2, 24		// timer->maxval = maxval
	std Z+MESSAGE_STRUCT_SIZE+3, 25

	std Z+0, 22		// timer->msg.pfunc
	std Z+1, 23
	std Z+2, 20		// timer->msg.msg
	std Z+3, 21
	std Z+4, 18		// timer->msg.param
	std Z+5, 19

	out _SFR_IO_ADDR(SREG), 0
	pop 16
	movw 24, ZL
	ret
.endfunc


.global ResetTimer		// int ResetTimer(HTIMER, unsigned int, void(*)(int, int), int, int);
.func ResetTimer
ResetTimer:
	cp 22, 1
	cpc 23, 1
	breq reset_count_null	// if(count==0) return -2

	ldi XL, hi8(_timers)
	cpi 24, lo8(_timers)
	cpc 25, XL
	brcs reset_wrong_handle
	ldi XL, hi8(_timers+(MAX_TIMERS*TIMER_STRUCT_SIZE))	// prfen, ob hthread auf einen Thread zeigt
	cpi 24, lo8(_timers+(MAX_TIMERS*TIMER_STRUCT_SIZE))
	cpc 25, XL
	brcc reset_wrong_handle

	movw ZL, 24	// Z = HTIMER

	cp 20, 1
	cpc 21, 1
	brne reset_timer_not_null

	lds 20, system_def_proc
	lds 21, system_def_proc+1

reset_timer_not_null:
	in 0, _SFR_IO_ADDR(SREG)
	cli

	std Z+MESSAGE_STRUCT_SIZE+0, 22		// timer->count = count
	std Z+MESSAGE_STRUCT_SIZE+1, 23
	std Z+MESSAGE_STRUCT_SIZE+2, 22		// timer->maxval = maxval
	std Z+MESSAGE_STRUCT_SIZE+3, 23

	std Z+0, 20		// timer->msg.pfunc
	std Z+1, 21
	std Z+2, 18		// timer->msg.msg
	std Z+3, 19
	std Z+4, 16		// timer->msg.param
	std Z+5, 17

	out _SFR_IO_ADDR(SREG), 0
	clr 24
	clr 25
	ret

reset_count_null:
	ldi 24, lo8(-2)
	ldi 25, hi8(-2)
	ret

reset_wrong_handle:
	ldi 24, lo8(-1)
	ldi 25, hi8(-1)
	ret
.endfunc

.global ResetAlarm		// int ResetAlarm(HTIMER, unsigned int, void(*)(int, int), int, int);
.func ResetAlarm
ResetAlarm:
	cp 22, 1
	cpc 23, 1
	breq reset_alarm_count_null	// if(count==0) return -2

	ldi XL, hi8(_timers)
	cpi 24, lo8(_timers)
	cpc 25, XL
	brcs reset_alarm_wrong_handle
	ldi XL, hi8(_timers+(MAX_TIMERS*TIMER_STRUCT_SIZE))	// prfen, ob hthread auf einen Thread zeigt
	cpi 24, lo8(_timers+(MAX_TIMERS*TIMER_STRUCT_SIZE))
	cpc 25, XL
	brcc reset_alarm_wrong_handle

	movw ZL, 24	// Z = HTIMER

	cp 20, 1
	cpc 21, 1
	brne reset_alarm_not_null

	lds 20, system_def_proc
	lds 21, system_def_proc+1

reset_alarm_not_null:
	in 0, _SFR_IO_ADDR(SREG)
	cli

	std Z+MESSAGE_STRUCT_SIZE+0, 22		// timer->count = count
	std Z+MESSAGE_STRUCT_SIZE+1, 23
	std Z+MESSAGE_STRUCT_SIZE+2, 1		// timer->maxval = 0
	std Z+MESSAGE_STRUCT_SIZE+3, 1

	std Z+0, 20		// timer->msg.pfunc
	std Z+1, 21
	std Z+2, 18		// timer->msg.msg
	std Z+3, 19
	std Z+4, 16		// timer->msg.param
	std Z+5, 17

	out _SFR_IO_ADDR(SREG), 0
	clr 24
	clr 25
	ret

reset_alarm_count_null:
	ldi 24, lo8(-2)
	ldi 25, hi8(-2)
	ret

reset_alarm_wrong_handle:
	ldi 24, lo8(-1)
	ldi 25, hi8(-1)
	ret
.endfunc

.global SetAlarm		// HTIMER SetAlarm(unsigned int, void(*)(int, int), int, int);
.func SetAlarm
SetAlarm:
	cp 24, 1
	cpc 25, 1
	breq alarm_no_count	// if(count==0) return NULL

	lds ZL, _timerindex
	lds ZH, _timerindex+1

	in 0, _SFR_IO_ADDR(SREG)
	cli

	push 16
	// nach einem freien Slot suchen
	ldi 16, hi8(_timers+(MAX_TIMERS*TIMER_STRUCT_SIZE))
while_alarm_thread:
	cpi ZL, lo8(_timers+(MAX_TIMERS*TIMER_STRUCT_SIZE))
	cpc ZH, 16
	brcc alarm_thread_no_free_slots
	ldd XL, Z+MESSAGE_STRUCT_SIZE+0
	ldd XH, Z+MESSAGE_STRUCT_SIZE+1
	cp XL, 1
	cpc XH, 1
	breq alarm_thread_slot_found		// while(timer->count)timer++
	adiw ZL, TIMER_STRUCT_SIZE
	rjmp while_timer_thread

alarm_thread_no_free_slots:
	sts _timerindex, ZL
	out _SFR_IO_ADDR(SREG), 0
	sts _timerindex+1, ZH
	pop 16
	clr 24
	clr 25
alarm_no_count:
	ret

alarm_thread_slot_found:

	cp 22, 1
	cpc 23, 1
	brne set_alarm_not_null

	lds 22, system_def_proc
	lds 23, system_def_proc+1

set_alarm_not_null:
	movw XL, ZL						// neuen timerindex speichern
	adiw XL, TIMER_STRUCT_SIZE
	sts _timerindex, XL
	sts _timerindex+1, XH

	std Z+MESSAGE_STRUCT_SIZE+0, 24		// timer->count = count
	std Z+MESSAGE_STRUCT_SIZE+1, 25
	std Z+MESSAGE_STRUCT_SIZE+2, 1		// timer->maxval = 0
	std Z+MESSAGE_STRUCT_SIZE+3, 1

	std Z+0, 22		// timer->msg.pfunc
	std Z+1, 23
	std Z+2, 20		// timer->msg.msg
	std Z+3, 21
	std Z+4, 18		// timer->msg.param
	std Z+5, 19

	out _SFR_IO_ADDR(SREG), 0
	pop 16
	movw 24, ZL
	ret
.endfunc


.global SetStopwatch // HTIMER SetStopwatch()
.func SetStopwatch
SetStopwatch:
	lds ZL, _timerindex
	lds ZH, _timerindex+1

	in 0, _SFR_IO_ADDR(SREG)
	cli

	push 16
	// nach einem freien Slot suchen
	ldi 16, hi8(_timers+(MAX_TIMERS*TIMER_STRUCT_SIZE))
while_setwatch:
	cpi ZL, lo8(_timers+(MAX_TIMERS*TIMER_STRUCT_SIZE))
	cpc ZH, 16
	brcc setwatch_no_free_slots
	ldd XL, Z+MESSAGE_STRUCT_SIZE+0
	ldd XH, Z+MESSAGE_STRUCT_SIZE+1
	cp XL, 1
	cpc XH, 1
	breq setwatch_slot_found		// while(timer->count)timer++
	adiw ZL, TIMER_STRUCT_SIZE
	rjmp while_setwatch

setwatch_no_free_slots:
	sts _timerindex, ZL
	out _SFR_IO_ADDR(SREG), 0
	sts _timerindex+1, ZH
	pop 16
	clr 24		// return NULL
	clr 25
	ret

setwatch_slot_found:
	movw XL, ZL							// neuen timerindex speichern
	adiw XL, TIMER_STRUCT_SIZE
	sts _timerindex, XL
	sts _timerindex+1, XH

	ldi 24, 0xFF

	std Z+MESSAGE_STRUCT_SIZE+0, 24		// timer->count = 0xFF
	std Z+MESSAGE_STRUCT_SIZE+1, 24
	std Z+MESSAGE_STRUCT_SIZE+2, 24		// timer->maxval = 0xFF
	std Z+MESSAGE_STRUCT_SIZE+3, 24

	std Z+0, 1		// timer->msg.pfunc
	std Z+1, 1
	std Z+2, 1		// timer->msg.msg
	std Z+3, 1

	out _SFR_IO_ADDR(SREG), 0
	pop 16
	movw 24, ZL
	ret
.endfunc


.global ResetStopwatch // int ResetStopwatch(HTIMER)
.func ResetStopwatch
ResetStopwatch:
	ldi XL, hi8(_timers)
	cpi 24, lo8(_timers)
	cpc 25, XL
	//brcs resetwatch_wrong_handle
	ldi XL, hi8(_timers+(MAX_TIMERS*TIMER_STRUCT_SIZE))	// prfen, ob hthread auf einen Thread zeigt
	cpi 24, lo8(_timers+(MAX_TIMERS*TIMER_STRUCT_SIZE))
	cpc 25, XL
	//brcc resetwatch_wrong_handle

	movw ZL, 24	// Z = HTIMER

	in 0, _SFR_IO_ADDR(SREG)
	cli

	ldi 24, 0xFF
	std Z+MESSAGE_STRUCT_SIZE+0, 24		// timer->count = count
	std Z+MESSAGE_STRUCT_SIZE+1, 24
	std Z+MESSAGE_STRUCT_SIZE+2, 24		// timer->maxval = maxval
	std Z+MESSAGE_STRUCT_SIZE+3, 24

	std Z+0, 1		// timer->msg.pfunc
	std Z+1, 1
	std Z+2, 1		// timer->msg.msg
	std Z+3, 1

	out _SFR_IO_ADDR(SREG), 0
	clr 24
	clr 25
	ret

resetwatch_wrong_handle:
	ldi 24, lo8(-1)
	ldi 25, hi8(-1)
	ret
.endfunc

.global SetResumeAlarm	// HTIMER SetResumeAlarm(unsigned int, HTHREAD);
.func SetResumeAlarm
SetResumeAlarm:
	cp 24, 1
	cpc 25, 1
	breq resume_alarm_no_count	// if(count==0) return NULL

	lds ZL, _timerindex
	lds ZH, _timerindex+1

	in 0, _SFR_IO_ADDR(SREG)
	cli

	// nach einem freien Slot suchen
	ldi 18, hi8(_timers+(MAX_TIMERS*TIMER_STRUCT_SIZE))
while_resume_alarm_thread:
	cpi ZL, lo8(_timers+(MAX_TIMERS*TIMER_STRUCT_SIZE))
	cpc ZH, 18
	brcc resume_alarm_thread_no_free_slots
	ldd XL, Z+MESSAGE_STRUCT_SIZE+0
	ldd XH, Z+MESSAGE_STRUCT_SIZE+1
	cp XL, 1
	cpc XH, 1
	breq resume_alarm_thread_slot_found		// while(timer->count)timer++
	adiw ZL, TIMER_STRUCT_SIZE
	rjmp while_resume_alarm_thread

resume_alarm_thread_no_free_slots:
	sts _timerindex, ZL
	out _SFR_IO_ADDR(SREG), 0
	sts _timerindex+1, ZH
resume_alarm_no_count:
	ret

resume_alarm_thread_slot_found:

	movw XL, ZL						// neuen timerindex speichern
	adiw XL, TIMER_STRUCT_SIZE
	sts _timerindex, XL
	sts _timerindex+1, XH

	std Z+MESSAGE_STRUCT_SIZE+0, 24		// timer->count = count
	std Z+MESSAGE_STRUCT_SIZE+1, 25
	std Z+MESSAGE_STRUCT_SIZE+2, 1		// timer->maxval = 0
	std Z+MESSAGE_STRUCT_SIZE+3, 1

	std Z+0, 1		// timer->msg.pfunc
	std Z+1, 1
	std Z+2, 22		// timer->msg.msg
	std Z+3, 23

	out _SFR_IO_ADDR(SREG), 0
	movw 24, ZL
	ret
.endfunc

.global KillTimer		// int KillTimer(HTIMER);
.func KillTimer
KillTimer:
	in 0, _SFR_IO_ADDR(SREG)
	cli

	ldi 23, hi8(_timers+(MAX_TIMERS*TIMER_STRUCT_SIZE))	// prfen, ob htimer auf einen Timer zeigt
	cpi 24, lo8(_timers+(MAX_TIMERS*TIMER_STRUCT_SIZE))
	cpc 25, 23
	brcc killtimer_wrong_handle
	ldi 23, hi8(_timers)
	cpi 24, lo8(_timers)
	cpc 25, 23
	brcs killtimer_wrong_handle

	movw ZL, 24

	std Z+MESSAGE_STRUCT_SIZE+0, 1 	// htimer->count = 0
	std Z+MESSAGE_STRUCT_SIZE+1, 1
	std Z+MESSAGE_STRUCT_SIZE+2, 1	// htimer->maxval = 0
	std Z+MESSAGE_STRUCT_SIZE+3, 1

	lds 24, _timerindex
	lds 25, _timerindex+1

	cp ZL, 24
	cpc ZH, 25
	brcc kill_no_timerindex

	sts _timerindex, ZL
	sts _timerindex+1, ZH

kill_no_timerindex:
	out _SFR_IO_ADDR(SREG), 0

	clr 24
	clr 25
	ret

killtimer_wrong_handle:
	out _SFR_IO_ADDR(SREG), 0
	ldi 24, lo8(-1)
	ldi 25, hi8(-1)
	ret
.endfunc


.global GetTimerCount	// unsigned int GetTimerCount(HTIMER);
.func GetTimerCount
GetTimerCount:
	ldi 23, hi8(_timers)
	cpi 24, lo8(_timers)
	cpc 25, 23
	brcs gettimer_wrong_handle
	ldi 23, hi8(_timers+(MAX_TIMERS*TIMER_STRUCT_SIZE))	// prfen, ob hthread auf einen Thread zeigt
	cpi 24, lo8(_timers+(MAX_TIMERS*TIMER_STRUCT_SIZE))
	cpc 25, 23
	brcc gettimer_wrong_handle

	in 0, _SFR_IO_ADDR(SREG)
	cli

	movw ZL, 24

	ldd 24, Z+MESSAGE_STRUCT_SIZE+0		// return htimer->count
	out _SFR_IO_ADDR(SREG), 0
	ldd 25, Z+MESSAGE_STRUCT_SIZE+1

	ret

gettimer_wrong_handle:
	clr 24
	clr 25
	ret
.endfunc

